
from __future__ import print_function
from parser import parse as _parse
from itertools import combinations
from nodes import NamedTerm

def expand_conditionals(rule, format = True):
  def all_subsets(set_):
    list_ = list(set_)
    for size in range(0, len(list_)+1):
      for subset in combinations(list_, size):
        yield subset

  ret = []
  if isinstance(rule, basestring):
    rule = _parse(rule)

  conditionals = rule.collect_conditionals()
  for subset in all_subsets(conditionals):
    # ensure that only terms in the subset are enabled
    if format:
      ret.append(rule.select(subset).format())
    else:
      ret.append(rule.select(subset))

  return sorted(set(ret))

def create_wrapper(rule, fn):
  if isinstance(rule, basestring):
    rule = _parse(rule)

  # flattening
  rule = rule.select(())

  def wrapper(p):
    kwargs = {}
    for i, term in enumerate(rule.terms):
      if isinstance(term, NamedTerm):
        kwargs[term.name] = p[i+1]
    p[0] = fn(**kwargs)
  
  wrapper.__doc__ = rule.format()

  return wrapper

def format(rules):
  if isinstance(rules, basestring):
    rules = parse(rules)

  return '\n'.join(rule.format() for rule in rules)

def parse(defn):
  defn = [line.strip() for line in defn.split('\n') if line.strip()]
  return [_parse(line) for line in defn]

def process_function(fn):
  rules = parse(fn.__doc__)
  expanded = sum((expand_conditionals(rule, format = False) for rule in rules), [])

  ret = {}
  i = 0
  for rule in expanded:
    ret['p_%s_%s' % (fn.__name__, i)] = create_wrapper(rule, fn)
    i += 1
  return ret

def process_all(globals, prefix = 'px_'):
  names = [var for var in globals.keys() if var.startswith(prefix)]
  funs = [globals[name] for name in names]
  for name in names:
    del globals[name]

  for fn in funs:
    for name, wrapper in process_function(fn).items():
      globals[name] = wrapper

  



from setuptools import setup

setup(name='easyply',
      version='1',
      description='Helper functions for PLY (Python Lex & Yacc)',
      author='Mike Hordecki',
      author_email='mike@hordecki.com',
      packages=['easyply'],
      long_description = open('README.rst').read(),
      install_requires=[
        "ply",
        ]
      )




class ComparisonMixin(object):
  @staticmethod
  def __get_cmpkey(obj):
    if not hasattr(obj, '_cmpkey'):
      raise NotImplementedError
    return obj._cmpkey()

  def __lt__(self, other):
    return ComparisonMixin.__get_cmpkey(self) < ComparisonMixin.__get_cmpkey(other)

  def __hash__(self):
    return hash(ComparisonMixin.__get_cmpkey(self))

  def __eq__(self, other):
    #print ComparisonMixin.__get_cmpkey(self) ,'==', ComparisonMixin.__get_cmpkey(other)
    return ComparisonMixin.__get_cmpkey(self) == ComparisonMixin.__get_cmpkey(other)

class Rule(ComparisonMixin):
  def __init__(self, name, terms):
    self.name = name
    self.terms = terms

  def __str__(self):
    return '%s: %s' % (self.name, self.terms)

  def format(self):
    return ' '.join((self.name, ':') + self.terms.format())

  def collect_conditionals(self):
    return self.terms.collect_conditionals()

  def select(self, enabled_conditionals):
    return Rule(self.name, Terms(self.terms.select(enabled_conditionals)))

  def _cmpkey(self):
    return (type(self), self.name, self.terms)

class Terms(ComparisonMixin):
  def __init__(self, terms):
    self.terms = tuple(terms)

  def append(self, term):
    self.terms += (term, )

  def __iter__(self):
    return iter(self.terms)

  def __str__(self):
    return ' '.join(str(term) for term in self.terms)

  def format(self):
    return sum((term.format() for term in self.terms), ())

  def collect_conditionals(self):
    return sum((term.collect_conditionals() for term in self.terms), ())

  def select(self, enabled_conditionals):
    return sum((term.select(enabled_conditionals) for term in self.terms), ())

  def _cmpkey(self):
    return (type(self), self.terms)

class ConditionalTerm(ComparisonMixin):
  def __init__(self, term):
    self.term = term

  def __str__(self):
    if isinstance(self.term, Terms):
      return '(%s)?' % self.term
    else:
      return str(self.term) + '?'

  def format(self):
    return self.term.format()

  def collect_conditionals(self):
    return (self, ) + self.term.collect_conditionals()

  def select(self, enabled_conditionals):
    if any(x is self for x in enabled_conditionals):
      return self.term.select(enabled_conditionals)
    else:
      return ()

  def _cmpkey(self):
    return (type(self), self.term)

class NamedTerm(ComparisonMixin):
  def __init__(self, parser_term, name):
    self.parser_term = parser_term
    self.name = name
  
  def __str__(self):
    return '{%s:%s}' % (self.parser_term, self.name)

  def format(self): return (self.parser_term, )
  def collect_conditionals(self): return ()
  def select(self, enabled_conditionals): return (self, )

  def _cmpkey(self):
    return (type(self), self.parser_term, self.name)

class Term(ComparisonMixin):
  def __init__(self, parser_term):
    self.parser_term = parser_term

  def format(self): return (self.parser_term, )
  def collect_conditionals(self): return ()
  def select(self, enabled_conditionals): return (self, )

  def __str__(self):
    return self.parser_term

  def _cmpkey(self):
    return (type(self), self.parser_term)





from __future__ import print_function

from ply.lex import lex
from ply.yacc import yacc

from nodes import Rule, Term, ConditionalTerm, NamedTerm, Terms


t_ID = r'[a-zA-Z_][a-zA-Z0-9_]*'
t_COLON = r':'

t_PAREN_BEGIN = r'\('
t_PAREN_END = r'\)'
t_CURLY_BEGIN = r'{'
t_CURLY_END = r'}'
t_QUESTION = r'\?'

def t_COMMENT(t):
  r'\#.*'
  return None

def t_error(a):
  assert False

t_ignore = ' \t'

tokens = []
name = None
tokens = [name[2:] for name in globals() if name.startswith('t_')]
tokens.remove('ignore')
tokens.remove('error')

def p_rule(p):
  "rule : rule_header terms"
  p[0] = Rule(p[1], p[2])

def p_rule_header(p):
  "rule_header : ID COLON"
  p[0] = p[1]
  
def p_terms(p):
  "terms : term"
  p[0] = Terms([p[1]])

def p_terms_append(p):
  "terms : terms term"
  p[0] = p[1]
  p[0].append(p[2])

def p_term(p):
  """
  term : conditional_term
  term : named_term
  """
  p[0] = p[1]

def p_term_literal(p):
  "term : ID"
  p[0] = Term(p[1])
  
def p_term_parens(p):
  "term : PAREN_BEGIN terms PAREN_END"
  p[0] = p[2]

def p_conditional_term(p):
  "conditional_term : term QUESTION"
  p[0] = ConditionalTerm(p[1])

def p_named_term(p):
  "named_term : CURLY_BEGIN ID COLON ID CURLY_END"
  p[0] = NamedTerm(p[2], p[4])

def p_named_term_default(p):
  "named_term : CURLY_BEGIN ID CURLY_END"
  p[0] = NamedTerm(p[2], p[2].lower())

def p_error(p):
  assert False, p

lexer = lex()
parser = yacc(picklefile = 'easyply.tab') #FIXME: rename parsetab

def parse(text):
  return parser.parse(text)

